package de.ozgcloud.eingang.xta;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.Collections;
import java.util.Iterator;
import java.util.stream.Stream;

import javax.xml.namespace.QName;
import javax.xml.transform.TransformerException;

import org.apache.commons.io.IOUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.boot.webservices.client.WebServiceTemplateBuilder;
import org.springframework.oxm.jaxb.Jaxb2Marshaller;
import org.springframework.stereotype.Service;
import org.springframework.ws.WebServiceMessage;
import org.springframework.ws.client.core.WebServiceMessageCallback;
import org.springframework.ws.client.core.WebServiceMessageExtractor;
import org.springframework.ws.soap.SoapFaultDetailElement;
import org.springframework.ws.soap.SoapHeader;
import org.springframework.ws.soap.SoapHeaderElement;
import org.springframework.ws.soap.SoapMessage;
import org.springframework.ws.soap.addressing.client.ActionCallback;
import org.springframework.ws.soap.addressing.version.Addressing10;
import org.springframework.ws.soap.client.SoapFaultClientException;
import org.springframework.ws.support.MarshallingUtils;
import org.w3._2005._08.addressing.AttributedURIType;

import de.ozgcloud.eingang.common.errorhandling.TechnicalException;
import de.xoev.transport.xta._211.ContentType;
import de.xoev.transport.xta._211.ExceptionType;
import de.xoev.transport.xta._211.GenericContentContainer;
import eu.osci.ws._2008._05.transport.MsgBoxCloseRequestType;
import eu.osci.ws._2008._05.transport.MsgBoxFetchRequest;
import eu.osci.ws._2008._05.transport.MsgBoxGetNextRequestType;
import eu.osci.ws._2008._05.transport.MsgBoxResponseType;
import eu.osci.ws._2008._05.transport.MsgBoxStatusListRequestType;
import eu.osci.ws._2008._05.transport.MsgSelector;
import eu.osci.ws._2008._05.transport.MsgStatusListType;
import eu.osci.ws._2008._05.transport.ObjectFactory;
import eu.osci.ws._2014._10.transport.MessageMetaData;
import jakarta.validation.Valid;
import jakarta.xml.bind.JAXBElement;
import lombok.NonNull;
import lombok.extern.log4j.Log4j2;

@Log4j2
@Service
class XtaRemoteService {

	private static final String ERROR_ON_CLOSE_LOG_TEMPLATE = "Error result on close request.\nReason: %s";
	private static final String DETAIL_LOG_TEMPLATE = "Code: %s, Message: %s";

	@Autowired
	@Valid
	private XtaProperties properties;

	@Autowired
	private XtaMessageMetaDataMapper mapper;

	@Autowired
	private WebServiceTemplateBuilder webServiceTemplateBuilder;

	@Autowired
	@Qualifier("osciTransportMarshaller")
	private Jaxb2Marshaller osciMarshaller;

	@Autowired
	@Qualifier("xoevTransportMarshaller")
	private Jaxb2Marshaller xoevMarshaller;

	public XtaMessageMetaDatasAndHeader getMessagesMetadata() {
		return mapper.msgStatusListFromSoap(getStatusList());
	}

	public XtaMessageMetaDatasAndHeader getNextMessagesMetadata(String msgBoxRequestId) {
		return mapper.msgStatusListFromSoap(getNextStatusList(msgBoxRequestId));
	}

	MsgStatusListTypeAndHeaderResponse getStatusList() {
		var request = buildListRequest();
		return getGenericStatusList(request);
	}

	MsgStatusListTypeAndHeaderResponse getNextStatusList(String msgBoxRequestId) {
		var request = buildNextListRequest(msgBoxRequestId);
		return getGenericStatusList(request);
	}

	MsgStatusListTypeAndHeaderResponse getGenericStatusList(Object request) {

		var template = webServiceTemplateBuilder.setMarshaller(osciMarshaller).setUnmarshaller(osciMarshaller).build();

		return template.sendAndReceive(buildMarshalCallBack(request, buildActionCallback()), buildHeaderExtractor());
	}

	private ActionCallback buildActionCallback() {
		return new ActionCallback(properties.getActions().getStatusList(), new Addressing10(), getTargetUri());
	}

	WebServiceMessageCallback buildMarshalCallBack(Object jaxbElement, ActionCallback callback) {
		return new WebServiceMessageCallback() {
			@Override
			public void doWithMessage(WebServiceMessage message) throws IOException, TransformerException {
				MarshallingUtils.marshal(osciMarshaller, jaxbElement, message);
				callback.doWithMessage(message);
			}
		};
	}

	WebServiceMessageExtractor<MsgStatusListTypeAndHeaderResponse> buildHeaderExtractor() {
		return new WebServiceMessageExtractor<MsgStatusListTypeAndHeaderResponse>() {
			@Override
			public MsgStatusListTypeAndHeaderResponse extractData(WebServiceMessage message) throws IOException, TransformerException {

				MsgBoxResponseType header = extractHeader(message);

				return MsgStatusListTypeAndHeaderResponse.builder()
						.msgBoxRequestID(header.getMsgBoxRequestID())
						.noMessageAvailable(header.getNoMessageAvailable() != null)
						.messageItemsPending(header.getItemsPending())
						.messages(extractMessages(message))
						.build();
			}

			@SuppressWarnings("unchecked")
			private MsgBoxResponseType extractHeader(WebServiceMessage message) {
				SoapHeader soapHeader = ((SoapMessage) message).getSoapHeader();
				Iterator<SoapHeaderElement> it = soapHeader
						.examineHeaderElements(new QName("http://www.osci.eu/ws/2008/05/transport", "MsgBoxResponse"));
				validateHasHeader(it);
				return ((JAXBElement<MsgBoxResponseType>) osciMarshaller.unmarshal(it.next().getSource())).getValue();
			}

			private void validateHasHeader(Iterator<SoapHeaderElement> it) {
				if (!it.hasNext()) {
					throw new TechnicalException("Response from XTA GetStatusList has no header");
				}
			}

			@SuppressWarnings("unchecked")
			private Stream<MessageMetaData> extractMessages(WebServiceMessage message) throws IOException {
				return ((JAXBElement<MsgStatusListType>) MarshallingUtils.unmarshal(osciMarshaller, message)).getValue().getMessageMetaData()
						.stream();
			}
		};
	}

	private JAXBElement<MsgBoxStatusListRequestType> buildListRequest() {
		ObjectFactory objectFactory = new ObjectFactory();

		MsgBoxStatusListRequestType msg = new MsgBoxStatusListRequestType();
		msg.setMaxListItems(properties.getMaxListElements());
		return objectFactory.createMsgBoxStatusListRequest(msg);
	}

	private JAXBElement<MsgBoxGetNextRequestType> buildNextListRequest(String msgBoxRequestId) {
		ObjectFactory objectFactory = new ObjectFactory();

		MsgBoxGetNextRequestType msg = new MsgBoxGetNextRequestType();
		msg.setMsgBoxRequestID(msgBoxRequestId);
		return objectFactory.createMsgBoxGetNextRequest(msg);
	}

	private URI getTargetUri() {
		try {
			return new URI(buildServerAddressUri());
		} catch (URISyntaxException e) {
			throw new TechnicalException("Error building target url: " + e);
		}
	}

	String buildServerAddressUri() {
		return XtaRemoteServiceConfiguration.URI_TEMPLATE.formatted(properties.getServer().getProtocol(),
				properties.getServer().getName());
	}

	public XtaMessage getMessage(XtaMessageId messageId) {
		return XtaMessage.builder()
				.metaData(null)
				.messageFiles(Collections.singleton(getMessage(messageId.toString())))
				.build();
	}

	XtaFile getMessage(String messageId) {
		var callback = new ActionCallback(properties.getActions().getFetchRequest(), new Addressing10(), getTargetUri());
		var template = webServiceTemplateBuilder.setMarshaller(osciMarshaller).setUnmarshaller(xoevMarshaller).build();

		var result = (GenericContentContainer) template.marshalSendAndReceive(buildFetchRequest(messageId), callback);
		return toXtaFile(result.getContentContainer().getMessage());
	}

	private XtaFile toXtaFile(ContentType type) {
		return XtaFile.builder()
				.file(persistToFile(type.getValue()))
				.name(type.getFilename())
				.contentType(type.getContentType())
				.size(type.getSize())
				.build();
	}

	private File persistToFile(byte[] data) {
		try {
			var file = File.createTempFile("xta", ".data");
			file.deleteOnExit();
			var out = new FileOutputStream(file);
			IOUtils.write(data, out);
			out.flush();
			out.close();
			return file;
		} catch (IOException e) {
			throw new TechnicalException("Error writing Attachment to temp file", e);
		}
	}

	private JAXBElement<MsgBoxFetchRequest> buildFetchRequest(String msgId) {
		MsgSelector msgSelector = new MsgSelector();
		AttributedURIType attribute = new AttributedURIType();
		attribute.setValue(msgId);
		msgSelector.getMessageID().add(attribute);

		var request = new MsgBoxFetchRequest();
		request.setMsgSelector(msgSelector);

		return wrapAsJaxBElemement(request);
	}

	private JAXBElement<MsgBoxFetchRequest> wrapAsJaxBElemement(MsgBoxFetchRequest request) {
		QName qname = new QName("http://www.osci.eu/ws/2008/05/transport", "MsgBoxFetchRequest");

		return new JAXBElement<>(qname, MsgBoxFetchRequest.class, request);
	}

	public void close(@NonNull XtaMessageId messageId) {
		var callback = new ActionCallback(properties.getActions().getCloseRequest(), new Addressing10(), getTargetUri());
		var template = webServiceTemplateBuilder.setMarshaller(osciMarshaller).setUnmarshaller(xoevMarshaller).build();

		try {
			template.marshalSendAndReceive(buildCloseRequest(messageId.toString()), callback);
		} catch (SoapFaultClientException e) {
			logErrorOnClose(e);
		}
	}

	private JAXBElement<MsgBoxCloseRequestType> buildCloseRequest(String msgId) {
		MsgBoxCloseRequestType request = new MsgBoxCloseRequestType();
		var lastMsgReceived = request.getLastMsgReceived();

		AttributedURIType attribute = new AttributedURIType();
		attribute.setValue(msgId);
		lastMsgReceived.add(attribute);

		return new ObjectFactory().createMsgBoxCloseRequest(request);
	}

	private void logErrorOnClose(SoapFaultClientException e) {
		try {
			var fault = e.getSoapFault();
			StringBuilder logBuilder = new StringBuilder(ERROR_ON_CLOSE_LOG_TEMPLATE.formatted(e.getSoapFault().getFaultStringOrReason()));

			var entries = fault.getFaultDetail().getDetailEntries();
			entries.forEachRemaining(entry -> logBuilder.append("\n").append(formatFaultEntry(entry)));

			LOG.error(logBuilder.toString(), e);
		} catch (Exception e1) {
			LOG.error("Error on loggging close error", e1);
			LOG.error("origin error was", e);
		}
	}

	private String formatFaultEntry(SoapFaultDetailElement soapfaultdetailelement1) {
		@SuppressWarnings("unchecked")
		ExceptionType exceptionType = ((JAXBElement<ExceptionType>) xoevMarshaller.unmarshal(soapfaultdetailelement1.getSource())).getValue();

		return DETAIL_LOG_TEMPLATE.formatted(exceptionType.getErrorCode().getCode(), exceptionType.getErrorCode().getName().toString());

	}

}
