package de.ozgcloud.eingang.xta;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.stream.Stream;

import org.assertj.core.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.vorgang.VorgangNummerSupplier;

class XtaServiceTest {

	private static final String MESSAGE_TYPE_OTHER = "MESSAGE_TYPE_OTHER";

	@Spy
	@InjectMocks
	private XtaService service;

	@Mock
	private XtaRemoteService remoteService;
	@Mock
	private XtaMessageMapper mapper;
	@Mock
	private VorgangNummerSupplier vorgangNummerSupplier;

	@Nested
	class TestGetMessagesAsFormData {

		private XtaMessageMetaDatasAndHeader metaData = XtaMessageMetaDatasAndHeaderTestFactory.create();
		private XtaMessageMetaData messageMetaData = XtaMessageMetaDataTestFactory.create();
		private XtaMessage message = XtaMessageTestFactory.create();

		@BeforeEach
		void setup() {
			doReturn(Stream.of(messageMetaData)).when(service).createXtaMessageStream();
		}

		@Test
		void shouldCallCreateStream() {
			service.getMessages();

			verify(service).createXtaMessageStream();
		}

		@Test
		void shouldCallFilterByMessageType() {
			when(mapper.toFormData(any())).thenReturn(FormDataTestFactory.create());
			when(remoteService.getMessage(any(XtaMessageId.class))).thenReturn(message);

			service.getMessages().toList();

			verify(service).filterByMessageType(messageMetaData);
		}

		@Test
		void shouldCallGetFormData() {
			when(mapper.toFormData(any())).thenReturn(FormDataTestFactory.create());
			when(remoteService.getMessage(any(XtaMessageId.class))).thenReturn(message);
			doReturn(true).when(service).filterByMessageType(messageMetaData);

			service.getMessages().toList();

			verify(service).getFormData(messageMetaData);
		}

		@Test
		void shouldNotCallGetFormData() {
			doReturn(false).when(service).filterByMessageType(messageMetaData);

			service.getMessages().toList();

			verify(service, never()).getFormData(any());
		}
	}

	@Nested
	class TestFilterByMessageType {

		@Test
		void shouldAcceptDFoerdermittel() {
			var metaDataDFoerder = XtaMessageMetaDataTestFactory.create();

			assertThat(service.filterByMessageType(metaDataDFoerder)).isTrue();
		}

		@Test
		void shouldNotAcceptOtherMessageType() {
			var metaDataDFoerder = XtaMessageMetaDataTestFactory.createBuilder().messageType(MESSAGE_TYPE_OTHER).build();

			assertThat(service.filterByMessageType(metaDataDFoerder)).isFalse();
		}
	}

	@Nested
	class TestGetFormData {

		private XtaMessage message = XtaMessageTestFactory.create();

		@Captor
		private ArgumentCaptor<XtaMessage> messageCaptor;
		@Mock
		private FormData formData;

		@BeforeEach
		void init() {
			when(remoteService.getMessage(any(XtaMessageId.class))).thenReturn(message);
		}

		@Test
		void shouldCallRemoteService() {
			doReturn(formData).when(service).updateHeader(any());

			service.getFormData(XtaMessageMetaDataTestFactory.create());

			verify(remoteService).getMessage(XtaMessageTestFactory.MESSAGE_ID);
		}

		@Test
		void shouldCallMapper() {
			doReturn(formData).when(service).updateHeader(any());

			service.getFormData(XtaMessageMetaDataTestFactory.create());

			verify(mapper).toFormData(any());
		}

		@Test
		void shouldHaveMetaData() {
			doReturn(formData).when(service).updateHeader(any());
			XtaMessageMetaData metaData = XtaMessageMetaDataTestFactory.create();

			service.getFormData(metaData);

			verify(mapper).toFormData(messageCaptor.capture());
			assertThat(messageCaptor.getValue().getMetaData()).isNotNull().isEqualTo(metaData);
		}

		@Test
		void shouldReturnMappedResult() {
			var mapped = FormDataTestFactory.create();
			when(mapper.toFormData(any())).thenReturn(mapped);
			doReturn(mapped).when(service).updateHeader(any());

			var result = service.getFormData(XtaMessageMetaDataTestFactory.create());

			assertThat(result).isSameAs(mapped);
		}
	}

	@Nested
	class TestUpdateHeader {

		@Test
		void shouldCallSetVorgangNummer() {
			var formData = FormDataTestFactory.create();

			service.updateHeader(formData);

			verify(service).setVorgangNummer(formData.getHeader());
		}

		@Test
		void shouldSetUpdatedHeader() {
			var updatedHeader = FormHeaderTestFactory.create();
			doReturn(updatedHeader).when(service).setVorgangNummer(any());

			var result = service.updateHeader(FormDataTestFactory.create());

			Assertions.assertThat(result.getHeader()).isSameAs(updatedHeader);
		}
	}

	@Nested
	class TestSetVorgangNummer {

		@Test
		void shouldCallVorgangNummerSupplier() {
			service.setVorgangNummer(FormHeaderTestFactory.create());

			verify(vorgangNummerSupplier).get(XtaService.VORGANG_NUMMER_SUFFIX_LENGTH);
		}

		@Test
		void shouldSetVorgangNummer() {
			var vorgangNummer = "vorgang-1";
			when(vorgangNummerSupplier.get(anyInt())).thenReturn(vorgangNummer);

			var result = service.setVorgangNummer(FormHeaderTestFactory.create());

			assertThat(result.getVorgangNummer()).isEqualTo(vorgangNummer);
		}
	}

	@Nested
	class TestAcknowledgeReceive {

		@Test
		void shouldCallRemoteService() {
			service.acknowledgeReceive(XtaMessageTestFactory.MESSAGE_ID);

			verify(remoteService).close(XtaMessageTestFactory.MESSAGE_ID);
		}
	}

}
